import asyncio as aio
import importlib.util
import inspect
import os
import pathlib
import sys
import traceback

import websockets.exceptions as wse

from ..meta import (
    EXIT_CLOSE,
    EXIT_RESTART,
    META_INFO,
    MODULE_MODE_FLAG,
    MODULE_MODE_SET,
)
from ..models.bot import BOT_PROXY, BotHookBus
from ..models.ipc import PluginBus, PluginStore
from ..models.plugin import Plugin
from ..types.core import IActionResponder
from ..types.exceptions import *
from ..types.models import BotLife
from ..types.typing import *
from ..types.utils import Logger
from ..utils.config import BotConfig
from ..utils.logger import get_logger
from .dispatcher import BotDispatcher
from .linker import BotLinker
from .responder import BotResponder

if sys.platform != "win32":
    import uvloop

    aio.set_event_loop_policy(uvloop.EventLoopPolicy())


class PluginLoader:
    """
    插件加载器
    """

    @classmethod
    def load_from_dir(cls, plugin_path: str) -> Plugin:
        """
        从指定插件目录加载插件
        """
        if not os.path.exists(os.path.join(plugin_path, "__init__.py")):
            raise PluginLoadError(
                f"{plugin_path} 缺乏入口主文件 __init__.py，插件无法加载"
            )
        plugin_name = os.path.basename(plugin_path)
        plugins_folder = str(pathlib.Path(plugin_path).parent.resolve(strict=True))
        plugins_folder_name = os.path.basename(plugins_folder)
        if plugins_folder not in sys.path:
            importlib.import_module(plugins_folder_name)
            sys.path.insert(0, plugins_folder)
        module = importlib.import_module(
            f"{plugins_folder_name}.{plugin_name}", f"{plugins_folder_name}"
        )

        plugin_class = None
        for obj in module.__dict__.values():
            if (
                isinstance(obj, type)
                and issubclass(obj, Plugin)
                and obj.__name__ != Plugin.__name__
            ):
                plugin_class = obj
                break
        if plugin_class is None:
            raise PluginLoadError(
                "指定的入口主文件中，未发现继承 Plugin 的插件类，无法加载插件"
            )
        plugin = plugin_class()
        file_path = inspect.getfile(module)
        return (plugin, file_path)

    @classmethod
    def load_from_type(cls, _class: Type[Plugin]) -> Plugin:
        """
        从插件类对象加载插件
        """
        plugin = _class()
        file_path = inspect.getfile(_class)
        return (plugin, file_path)

    @classmethod
    def load(
        cls,
        plugin_target: Union[str, Type[Plugin]],
        logger: Logger,
        responder: IActionResponder,
    ) -> Plugin:
        """
        加载插件
        """
        if isinstance(plugin_target, str):
            plugin, file_path = cls.load_from_dir(plugin_target)
        else:
            plugin, file_path = cls.load_from_type(plugin_target)
        root_path = pathlib.Path(file_path).parent.resolve(strict=True)
        plugin._Plugin__build(root_path, logger, responder)
        return plugin


class MeloBot:
    """
    bot 模块。负责整个 bot 的启动、加载和管理。
    指定 restart_tip 可在重启时自动回复
    """

    def __init__(self) -> None:
        self.config: BotConfig
        self.info = META_INFO
        # 不要更改这个属性名
        self._logger: Logger = None

        self.life: aio.Task = None
        self.plugins: Dict[str, Plugin] = {}
        self.loader = PluginLoader

        self.linker: BotLinker
        self.responder: BotResponder
        self.dispatcher: BotDispatcher
        self.plugin_bus = PluginBus
        self.plugin_store = PluginStore
        self.bot_bus = BotHookBus
        self.proxy = BOT_PROXY

        self.__init_flag__ = False
        self.exit_code = EXIT_CLOSE

    @property
    def loop(self) -> aio.AbstractEventLoop:
        return aio.get_running_loop()

    @property
    def slack(self) -> bool:
        return self.linker.slack

    @slack.setter
    def slack(self, value: bool) -> None:
        self.linker.slack = value

    @property
    def logger(self) -> Logger:
        return self._logger if self._logger else get_logger()

    def init(self, config_dir) -> None:
        """
        为 bot 核心加载配置文件
        """
        if self.__init_flag__:
            self.logger.error("bot 不能重复初始化")
            exit(self.exit_code)

        self.config = BotConfig(config_dir)
        self._logger = get_logger(self.config.log_dir_path, self.config.log_level)
        self.linker = BotLinker(
            self.config.connect_host,
            self.config.connect_port,
            self.config.max_conn_try,
            self.config.conn_try_interval,
            self.config.cooldown_time,
            self.logger,
        )
        self.responder = BotResponder(self.logger)
        self.dispatcher = BotDispatcher(self.logger)
        self.plugin_bus._bind(self.logger, self.responder)
        self.bot_bus._bind(self.logger, self.responder)

        self.logger.info("欢迎使用 melobot v2")
        self.logger.info(
            f"melobot 在 AGPL3 协议下开源发行。更多请参阅：{self.info.PROJ_SRC}"
        )
        self.logger.info(f"运行版本：{self.info.VER}，平台：{self.info.PLATFORM}")
        if os.environ.get(MODULE_MODE_FLAG) == MODULE_MODE_SET:
            self.logger.info("当前运行模式：模块运行模式")
        else:
            self.logger.info("当前运行模式：脚本运行模式")
        self.logger.info("bot 核心初始化完成")
        self.__init_flag__ = True

    def load_plugin(self, plugin_target: Union[str, Type[Plugin]]) -> None:
        """
        为 bot 加载运行插件。支持传入插件起始目录字符串（绝对路径）或插件类
        """
        if not self.__init_flag__:
            self.logger.error("加载插件必须在初始化之后进行")
            exit(self.exit_code)

        plugin_dir = (
            inspect.getfile(plugin_target)
            if not isinstance(plugin_target, str)
            else plugin_target
        )
        self.logger.debug(f"尝试加载来自 {plugin_dir} 的插件")
        plugin = self.loader.load(plugin_target, self.logger, self.responder)
        exist_plugin = self.plugins.get(plugin.__class__.__id__)
        if exist_plugin is None:
            self.plugins[plugin.__class__.__id__] = plugin
            self.dispatcher.add_handlers(plugin._Plugin__handlers)
            self.logger.info(f"成功加载插件：{plugin.__class__.__id__}")
        else:
            self.logger.error(
                f"加载插件出错：插件名称重复, 尝试加载：{plugin_dir}，已加载：{exist_plugin.plugin_dir}"
            )
            exit(self.exit_code)

    def load_plugins(self, plugins_dir: str) -> None:
        """
        从插件目录加载多个 bot 运行插件。但必须遵循标准插件格式
        """
        items = os.listdir(plugins_dir)
        for item in items:
            path = os.path.join(plugins_dir, item)
            if os.path.isdir(path) and os.path.basename(path) != "__pycache__":
                self.load_plugin(path)

    async def _run(self) -> None:
        if len(self.plugins) == 0:
            self.logger.warning("没有加载任何插件，bot 将不会有任何操作")

        await self.bot_bus.emit(BotLife.LOADED)
        self.responder.bind(self.linker)
        self.linker.bind(self.dispatcher, self.responder)
        try:
            async with self.linker:
                self.life = aio.create_task(self.linker.listen())
                self.proxy._bind(self)
                self.logger.info("bot 开始正常运行")
                await self.life
        except wse.ConnectionClosed:
            pass
        except Exception as e:
            self.logger.error(f"bot 核心无法继续运行。异常：{e}")
            self.logger.debug("异常回溯栈：\n" + traceback.format_exc().strip("\n"))
        finally:
            await self.bot_bus.emit(BotLife.BEFORE_STOP, wait=True)
            self.logger.info("bot 已清理运行时资源并关闭")
            exit(self.exit_code)

    def run(self) -> None:
        """
        运行 bot
        """
        if not self.__init_flag__:
            self.logger.error("必须先初始化才能启动")
            exit(self.exit_code)
        try:
            """
            一定要手动设置事件循环。在启动前，部分核心模块初始化异步对象，
            已经生成了事件循环，不设置而直接使用 asyncio.run 将会运行一个新的事件循环，
            不同事件循环的异步对象无法直接通信
            """
            aio.set_event_loop(aio.get_event_loop())
            # 使用 asyncio.run 可以保证发生各种异常时一定取消所有任务（包括键盘中断）
            aio.run(self._run())
        except KeyboardInterrupt:
            pass

    async def close(self) -> None:
        """
        停止 bot
        """
        if self.life is None:
            self.logger.error("bot 尚未运行，无需停止")
            exit(self.exit_code)
        await self.bot_bus.emit(BotLife.BEFORE_CLOSE, wait=True)
        self.life.cancel()

    async def restart(self) -> None:
        """
        重启 bot
        """
        if self.life is None:
            self.logger.error("bot 尚未运行，无需重启")
            exit(self.exit_code)
        if os.environ.get(MODULE_MODE_FLAG) != MODULE_MODE_SET:
            self.logger.error("只有在模块运行模式下，才能使用 bot 重启功能")
            return
        self.logger.info("bot 即将进行重新启动...")
        self.exit_code = EXIT_RESTART
        self.life.cancel()



import asyncio as aio
import time
import traceback
from itertools import count

import websockets
import websockets.exceptions as wse

from ..models.action import BotAction
from ..models.bot import BotHookBus
from ..models.event import BotEventBuilder
from ..types.core import IActionSender, IEventDispatcher, IRespDispatcher
from ..types.exceptions import *
from ..types.models import BotLife
from ..types.typing import *
from ..utils.logger import Logger


class BotLinker(IActionSender):
    """
    Bot 连接模块通过连接适配器的代理，完成事件接收与行为发送。
    """

    def __init__(
        self,
        connect_host: str,
        connect_port: int,
        max_retry: int,
        retry_delay: int,
        send_interval: float,
        logger: Logger,
    ) -> None:
        super().__init__()
        self.url = f"ws://{connect_host}:{connect_port}"
        self.conn = None

        self.logger = logger
        self.slack = False
        self.max_retry_num = max_retry
        self.retry_delay = retry_delay if retry_delay > 0 else 0

        self._ready_signal = aio.Event()
        self._send_lock = aio.Lock()
        self._rest_time = send_interval
        self._pre_send_time = time.time()
        self._common_dispatcher: IEventDispatcher
        self._resp_dispatcher: IRespDispatcher

    def bind(
        self,
        common_dispatcher: IEventDispatcher,
        resp_dispatcher: IRespDispatcher,
    ) -> None:
        """
        绑定其他核心组件的方法。
        """
        self._common_dispatcher = common_dispatcher
        self._resp_dispatcher = resp_dispatcher
        self._ready_signal.set()

    async def _start(self) -> None:
        """
        启动连接
        """
        iterator = count(0) if self.max_retry_num < 0 else range(self.max_retry_num + 1)
        for _ in iterator:
            try:
                self.conn = await websockets.connect(self.url)
                await self.conn.recv()
                self.logger.info("与连接适配器，建立了 ws 连接")
                await BotHookBus.emit(BotLife.CONNECTED)
                return
            except Exception as e:
                self.logger.warning(
                    f"ws 连接建立失败，{self.retry_delay}s 后自动重试。错误：{e}"
                )
                await aio.sleep(self.retry_delay)
        raise BotConnectFailed("连接重试已达最大重试次数，已放弃建立连接")

    async def _close(self) -> None:
        """
        关闭连接
        """
        await self.conn.close()
        self.logger.info("已经关闭与连接适配器的连接")

    async def __aenter__(self) -> "BotLinker":
        await self._start()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_val: str, exc_tb: traceback
    ) -> None:
        if exc_type == wse.ConnectionClosedError:
            self.logger.warning("连接适配器主动关闭, bot 将自动清理资源后关闭")
        await self._close()

    async def send(self, action: BotAction) -> None:
        """
        发送一个 action 给连接适配器
        """
        await self._ready_signal.wait()
        await BotHookBus.emit(BotLife.ACTION_PRESEND, action, wait=True)
        if self.slack:
            return

        try:
            async with self._send_lock:
                action_str = action.flatten()
                await aio.sleep(self._rest_time - (time.time() - self._pre_send_time))
                await self.conn.send(action_str)
                self._pre_send_time = time.time()
        except aio.CancelledError:
            self.logger.warning("发生一次递归重发")
            await self.send(action)
        except wse.ConnectionClosed:
            self.logger.error("与连接适配器的通信已经断开，无法再执行操作")

    async def listen(self) -> None:
        """
        从连接适配器接收一个事件，并转化为 BotEvent 对象传递给 dispatcher 处理
        """
        await self._ready_signal.wait()

        try:
            while True:
                try:
                    raw_event = await self.conn.recv()
                    if raw_event == "":
                        continue
                    event = BotEventBuilder.build(raw_event)
                    if event.is_resp():
                        aio.create_task(self._resp_dispatcher.dispatch(event))
                    else:
                        aio.create_task(self._common_dispatcher.dispatch(event))
                except wse.ConnectionClosed:
                    raise
                except Exception as e:
                    self.logger.error(
                        f"bot life_task 抛出异常：[{e.__class__.__name__}] {e}"
                    )
                    self.logger.debug(f"异常点的事件记录为：{raw_event}")
                    self.logger.debug(
                        "异常回溯栈：\n" + traceback.format_exc().strip("\n")
                    )
        except aio.CancelledError:
            self.logger.debug("bot 运行例程已停止")



import json
import re
import time

from ..types.models import BotEvent
from ..types.typing import *
from .base import to_cq_arr


class BotEventBuilder:
    @classmethod
    def build(cls, rawEvent: Union[dict, str]) -> BotEvent:
        if isinstance(rawEvent, str):
            rawEvent = json.loads(rawEvent)

        etype = rawEvent.get("post_type")
        if etype in ("message_sent", "message"):
            return MsgEvent(rawEvent)
        elif etype == "request":
            return RequestEvent(rawEvent)
        elif etype == "notice":
            return NoticeEvent(rawEvent)
        elif etype == "meta_event":
            return MetaEvent(rawEvent)
        else:
            return RespEvent(rawEvent)


class MsgEvent(BotEvent):
    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        self.bot_id = rawEvent.get("self_id")

        self.id: int
        self.sender: MsgEvent.Sender
        self.group_id: Union[int, None]
        # 使用 CQ 字符串编码的消息
        self.raw_content: str
        # array 格式表示所有类型消息段
        self.content: dict
        # 消息中所有文本消息段的合并字符串
        self.text: str
        self.font: int
        self.temp_src: Union[str, None]

        self._init()

    @property
    def time(self) -> int:
        return self.raw.get("time")

    @property
    def type(self) -> str:
        return "message"

    def _init(self) -> None:
        rawEvent = self.raw

        self._cq_regex = re.compile(r"\[CQ:.*?\]")
        self.id = rawEvent["message_id"]
        self.raw_content = rawEvent["raw_message"]
        self.content = self._format_to_array(rawEvent["message"])
        self.text = self._get_text(self.content)
        self.font = rawEvent["font"]

        self.temp_src = None
        temp_src = rawEvent.get("temp_source")
        if temp_src:
            self.temp_src = MsgEvent._TEMP_SRC_MAP[temp_src]

        self.sender = MsgEvent.Sender(
            rawEvent=rawEvent,
            isGroup=self.is_group(),
            isGroupAnonym=self.is_group_anonym(),
        )

        self.group_id = None
        if self.is_group():
            self.group_id = rawEvent["group_id"]

    def _format_to_array(self, content: Union[dict, str]) -> dict:
        if not isinstance(content, str):
            return content
        else:
            return to_cq_arr(content)

    def _get_text(self, content: Union[dict, str]) -> bool:
        """
        获取消息中所有文本消息，返回合并字符串
        """
        text_list = []
        for item in content:
            if item["type"] == "text":
                text_list.append(item["data"]["text"])
        return "".join(text_list)

    def cq_get(self, cq_type: str, param: str) -> Union[List[Union[int, float, str]], None]:
        """
        从当前 event 中获取指定 cq 类消息的指定 param，以列表形式返回。
        当没有对应类型的 cq 消息或者没有对应的 param，不会返回空列表，而是返回 None
        """
        if isinstance(self.content, str):
            pass
        else:
            results = []
            for item in self.content:
                if item["type"] == cq_type:
                    val = item["data"].get(param)
                    if val is not None:
                        results.append(val)
            return results if len(results) > 0 else None

    def is_private(self) -> bool:
        """是否为私聊消息（注意群临时会话属于该类别）"""
        return self.raw["message_type"] == "private"

    def is_friend(self) -> bool:
        """是否为好友消息"""
        return (
            self.raw["message_type"] == "private" and self.raw["sub_type"] == "friend"
        )

    def is_group(self) -> bool:
        """是否为群消息（正常群消息、群匿名消息、群自身消息、群系统消息属于该类型）"""
        return self.raw["message_type"] == "group"

    def is_group_normal(self) -> bool:
        """是否为正常群消息"""
        return self.raw["message_type"] == "group" and self.raw["sub_type"] == "normal"

    def is_group_anonym(self) -> bool:
        """是否为匿名群消息"""
        return (
            self.raw["message_type"] == "group" and self.raw["sub_type"] == "anonymous"
        )

    def is_group_self(self) -> bool:
        """是否为群自身消息（即 bot 自己群中发的消息）"""
        return (
            self.raw["message_type"] == "group" and self.raw["sub_type"] == "group_self"
        )

    def is_group_temp(self) -> bool:
        """是否为群临时会话（属于私聊的一种）"""
        return self.raw["message_type"] == "private" and self.raw["sub_type"] == "group"

    def is_temp(self) -> bool:
        """是否为临时会话（属于私聊的一种）"""
        return "temp_source" in self.raw.keys()

    def is_group_notice(self) -> bool:
        """是否为群系统消息，特别注意是消息！"""
        return self.raw["message_type"] == "group" and self.raw["sub_type"] == "notice"

    _TEMP_SRC_MAP = {
        0: "群聊",
        1: "QQ咨询",
        2: "查找",
        3: "QQ电影",
        4: "热聊",
        6: "验证消息",
        7: "多人聊天",
        8: "约会",
        9: "通讯录",
    }

    class Sender:
        """
        消息发送者信息类
        """

        def __init__(self, rawEvent: dict, isGroup: bool, isGroupAnonym: bool) -> None:
            self._rawEvent = rawEvent
            self._isGroup = isGroup
            self._isGroupAnonym = isGroupAnonym
            self.id: int
            self.nickname: str
            self.sex: str
            self.age: int

            self.group_card: str
            # 总共有四种：owner, admin, member, anonymous
            self.group_role: Literal["owner", "admin", "member", "anonymous"]
            self.group_title: str
            self.group_area: str
            self.group_level: str

            self.anonym_id: int
            self.anonym_name: str
            # 匿名用户 flag，调用 go-cqhttp 相关 API 时，需要使用
            self.anonym_flag: str

            # 对应私聊可能缺失 sender 的情况
            if len(rawEvent["sender"]) == 0:
                self._gen_empty()
                return

            self.id = rawEvent["sender"]["user_id"]
            self.nickname = rawEvent["sender"]["nickname"]
            self.sex = (
                rawEvent["sender"]["sex"] if "sex" in rawEvent["sender"].keys() else ""
            )
            self.age = (
                rawEvent["sender"]["age"] if "age" in rawEvent["sender"].keys() else ""
            )
            if isGroup:
                if isGroupAnonym:
                    self.group_card = ""
                    self.group_area = rawEvent["sender"]["area"]
                    self.group_level = rawEvent["sender"]["level"]
                    self.group_role = "anonymous"
                    self.group_title = ""
                    self.anonym_id = rawEvent["anonymous"]["id"]
                    self.anonym_name = rawEvent["anonymous"]["name"]
                    self.anonym_flag = rawEvent["anonymous"]["flag"]
                else:
                    self.group_card = rawEvent["sender"]["card"]
                    self.group_area = (
                        rawEvent["sender"]["area"]
                        if "area" in rawEvent["sender"].keys()
                        else ""
                    )
                    self.group_level = rawEvent["sender"]["level"]
                    self.group_role = rawEvent["sender"]["role"]
                    self.group_title = rawEvent["sender"]["title"]

        def _gen_empty(self) -> None:
            """当传入的 rawEvent 缺失 sender 时，把属性直接置空"""
            self.id = self._rawEvent["user_id"]
            self.nickname = None
            self.sex = None
            self.age = None

        def is_group_owner(self) -> bool:
            """判断是否为群主，若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "owner"

        def is_group_admin(self) -> bool:
            """判断是否为群管理（包含群主），若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "admin" or self.group_role == "owner"

        def only_group_member(self) -> bool:
            """判断是否只是群员（注意只是群员，不包括群主、管理和匿名），若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "member"

        def is_anonym_member(self) -> bool:
            """判断是否是群匿名，若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "anonymous"

        def is_bot(self) -> bool:
            """判断消息是否是bot自己发送的"""
            return self.id == self._rawEvent["self_id"]


class RequestEvent(BotEvent):
    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        self.bot_id = rawEvent.get("self_id")

        self.from_id: int
        self.from_group_id: Union[int, None]
        # 此处为加群或加好友的验证消息
        self.req_comment: str
        # 请求 flag，调用相关 go-cqhttp API 时，需要使用
        self.req_flag: str
        # 当为加群请求时，类型有：add, invite（加群请求和邀请 bot 入群）
        self.group_req_type: Union[Literal["add", "invite"], None]

        self._init()

    @property
    def time(self) -> int:
        return self.raw.get("time")

    @property
    def type(self) -> str:
        return "request"

    def _init(self) -> None:
        rawEvent = self.raw
        self.group_req_type = None
        self.from_group_id = None

        if self.is_friend_req():
            self.from_id = rawEvent["user_id"]
            self.req_comment = rawEvent["comment"]
            self.req_flag = rawEvent["flag"]
        elif self.is_group_req():
            self.group_req_type = rawEvent["sub_type"]
            self.from_id = rawEvent["user_id"]
            self.from_group_id = rawEvent["group_id"]
            self.req_comment = rawEvent["comment"]
            self.req_flag = rawEvent["flag"]

    def is_friend_req(self) -> bool:
        """是否为加好友请求"""
        return self.raw["request_type"] == "friend"

    def is_group_req(self) -> bool:
        """是否为加群请求"""
        return self.raw["request_type"] == "group"


class NoticeEvent(BotEvent):
    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        self.bot_id = rawEvent.get("self_id")
        # 修复某些 onebot 协议实现，user_id 缺失的问题
        if "target_id" in rawEvent.keys() and "user_id" not in rawEvent:
            rawEvent["user_id"] = rawEvent["target_id"]

        # 通知作用者或主体方的 id，如被禁言的一方
        self.notice_user_id: int
        # 通知若发生在群中，群 id
        self.notice_group_id: int
        # 通知发起者或操作方的 id，如禁言别人的管理员
        self.notice_operator_id: int
        # 通知涉及消息时，消息 id
        self.notice_msg_id: int
        # 若为入群通知，类型有：approve, invite（管理员同意和管理员邀请）
        self.join_group_type: str
        # 若为退群通知，类型有：leave, kick, kick_me
        self.leave_group_type: str
        # 若为群管理员变动通知，类型有：set, unset
        self.admin_change_type: str
        # 若为文件上传通知，此处为 通知文件信息对象
        self.file: NoticeEvent.File
        # 若为群禁言通知，类型有：ban, lift_ban
        self.group_ban_type: str
        # 若为禁言通知，禁言时长
        self.ban_time: int
        # 若为群荣誉变更通知，类型有：talkactive, performer, emotion
        self.honor_change_type: str
        # 若为群头衔变更通知，新头衔
        self.new_title: str
        # 若为群名片更新事件，旧名片和新名片；注意名片为空时，对应属性为空字符串
        self.old_card: str
        self.new_card: str
        # 若为客户端在线状态变更事件，此处为 通知客户端信息对象
        self.client: NoticeEvent.Client
        # 若为精华消息变更事件，类型有：add, delete
        self.essence_change_type: str

        self._init()

    @property
    def time(self) -> int:
        return self.raw.get("time")

    @property
    def type(self) -> str:
        return "notice"

    def _init(self) -> None:
        """
        外部确认为该类型事件时，调用此方法。
        """
        rawEvent = self.raw

        if self.is_friend_recall():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_msg_id = rawEvent["message_id"]
        elif self.is_group_recall():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.notice_msg_id = rawEvent["message_id"]
        elif self.is_group_increase():
            self.join_group_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
        elif self.is_group_decrease():
            self.leave_group_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
        elif self.is_group_admin():
            self.admin_change_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_group_upload():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.file = NoticeEvent.File(rawEvent, isGroup=True)
        elif self.is_group_ban():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.group_ban_type = rawEvent["sub_type"]
            self.ban_time = rawEvent["duration"]
        elif self.is_friend_add():
            self.notice_user_id = rawEvent["user_id"]
        elif self.is_poke():
            self.notice_user_id = rawEvent["target_id"]
            self.notice_operator_id = rawEvent["user_id"]
            if "group_id" in rawEvent.keys():
                self.notice_group_id = rawEvent["group_id"]
        elif self.is_lucky_king():
            self.notice_user_id = rawEvent["target_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["user_id"]
        elif self.is_honor():
            self.honor_change_type = rawEvent["honor_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_title():
            self.new_title = rawEvent["title"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_group_card():
            self.old_card = rawEvent["card_old"]
            self.new_card = rawEvent["card_new"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_offline_file():
            self.notice_user_id = rawEvent["user_id"]
            self.file = NoticeEvent.File(rawEvent, isGroup=False)
        elif self.is_client_status():
            self.client = NoticeEvent.Client(rawEvent)
        elif self.is_essence():
            self.essence_change_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["sender_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.notice_msg_id = rawEvent["message_id"]

    def is_group(self) -> bool:
        return "group_id" in self.raw.keys()

    def is_group_upload(self) -> bool:
        return self.raw["notice_type"] == "group_upload"

    def is_group_admin(self) -> bool:
        return self.raw["notice_type"] == "group_admin"

    def is_group_decrease(self) -> bool:
        return self.raw["notice_type"] == "group_decrease"

    def is_group_increase(self) -> bool:
        return self.raw["notice_type"] == "group_increase"

    def is_group_ban(self) -> bool:
        return self.raw["notice_type"] == "group_ban"

    def is_friend_add(self) -> bool:
        return self.raw["notice_type"] == "friend_add"

    def is_group_recall(self) -> bool:
        return self.raw["notice_type"] == "group_recall"

    def is_friend_recall(self) -> bool:
        return self.raw["notice_type"] == "friend_recall"

    def is_group_card(self) -> bool:
        return self.raw["notice_type"] == "group_card"

    def is_offline_file(self) -> bool:
        """是否为离线文件上传事件（即私聊文件上传）"""
        return self.raw["notice_type"] == "offline_file"

    def is_client_status(self) -> bool:
        return self.raw["notice_type"] == "client_status"

    def is_essence(self) -> bool:
        return self.raw["notice_type"] == "essence"

    def is_notify(self) -> bool:
        """是否为系统通知事件（包含群荣誉变更、戳一戳、群红包幸运王、群成员头衔变更）"""
        return self.raw["notice_type"] == "notify"

    def is_honor(self) -> bool:
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "honor"

    def is_poke(self) -> bool:
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "poke"

    def is_lucky_king(self) -> bool:
        return (
            self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "lucky_king"
        )

    def is_title(self) -> bool:
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "title"

    class File:
        """
        通知中文件信息类
        """

        def __init__(self, rawEvent: dict, isGroup: bool) -> None:
            self.id: str = None
            self.name: str = None
            self.size: int = None
            self.busid: int = None
            self.url: str = None

            self.name = rawEvent["file"]["name"]
            self.size = rawEvent["file"]["size"]
            if isGroup:
                self.id = rawEvent["file"]["id"]
                self.busid = rawEvent["file"]["busid"]
            else:
                self.url = rawEvent["file"]["url"]

    class Client:
        """
        通知中客户端信息类
        """

        def __init__(self, rawEvent: dict) -> None:
            self.online = rawEvent["online"]
            self.id = rawEvent["client"]["app_id"]
            self.name = rawEvent["client"]["device_name"]
            self.kind = rawEvent["client"]["device_kind"]


class MetaEvent(BotEvent):
    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        self.bot_id = rawEvent.get("self_id")

    @property
    def time(self) -> int:
        return self.raw.get("time")

    @property
    def type(self) -> str:
        return "meta"

    def is_lifecycle(self) -> bool:
        return self.raw["meta_event_type"] == "lifecycle"

    def is_heartbeat(self) -> bool:
        return self.raw["meta_event_type"] == "heartbeat"


class RespEvent(BotEvent):
    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        self.status = rawEvent.get("retcode")

        # 响应标识符
        self.id: str = None
        # 状态码
        self.status: int = None
        # 错误
        self.err: str = None
        # 错误提示
        self.err_prompt: str = None
        # 响应数据
        self.data: dict = None

        self._init()

    @property
    def time(self) -> int:
        return int(time.time())

    @property
    def type(self) -> str:
        return "response"

    def _init(self) -> None:
        rawEvent = self.raw
        self.status = rawEvent["retcode"]
        if "echo" in rawEvent.keys() and rawEvent["echo"]:
            self.id = rawEvent["echo"]
        if "data" in rawEvent.keys() and rawEvent["data"]:
            self.data = rawEvent["data"]

    def is_ok(self) -> bool:
        """判断是否为成功响应"""
        return self.raw["status"] == "ok"

    def is_processing(self) -> bool:
        """判断响应是否在被异步处理，即未完成但在处理中"""
        return self.status == 202

    def is_failed(self) -> bool:
        """判断是否为失败响应"""
        return self.raw["status"] != "ok"



import asyncio as aio
import traceback

from ..types.core import IActionResponder
from ..types.exceptions import *
from ..types.models import BotEvent, SessionRule
from ..types.typing import *
from ..types.utils import BotChecker, BotMatcher, BotParser, Logger
from .action import msg_action
from .event import MsgEvent
from .session import SESSION_LOCAL, BotSession, BotSessionManager


class EventHandler:
    def __init__(
        self,
        executor: AsyncFunc[None],
        plugin: Any,
        responder: IActionResponder,
        logger: Logger,
        checker: BotChecker,
        priority: PriorityLevel,
        timeout: float,
        set_block: bool,
        temp: bool,
        session_rule: SessionRule,
        session_hold: bool,
        direct_rouse: bool,
        conflict_wait: bool,
        conflict_cb_maker: Callable[[None], Coroutine],
        overtime_cb_maker: Callable[[None], Coroutine],
    ) -> None:
        super().__init__()
        self.is_valid = True

        self.executor = executor
        self.responder = responder
        self.logger = logger
        self.checker = checker
        self.priority = priority
        self.timeout = timeout
        self.set_block = set_block
        self.temp = temp

        self._run_lock = aio.Lock()
        self._rule = session_rule
        self._hold = session_hold
        self._plugin = plugin
        self._direct_rouse = direct_rouse
        self._conflict_cb_maker = conflict_cb_maker
        self._overtime_cb_maker = overtime_cb_maker
        self._wait_flag = conflict_wait

        if session_rule is None:
            if session_hold or direct_rouse or conflict_wait or conflict_cb_maker:
                raise EventHandlerError(
                    "使用 session_rule 参数后才能使用以下参数：session_hold， direct_rouse, \
                                      conflict_wait, conflict_callback"
                )

        if conflict_wait and conflict_cb_maker:
            raise EventHandlerError(
                "参数 conflict_wait 为 True 时，冲突回调永远不会被调用"
            )

    def _invalidate(self) -> None:
        """
        标记此 handler 为无效状态
        """
        self.is_valid = False

    def _verify(self, event: BotEvent) -> bool:
        """
        验证事件是否有触发执行的资格（验权）
        """
        if self.checker:
            return self.checker.check(event)
        return True

    async def _run_on_ctx(
        self, coro: Coroutine, session: BotSession = None, timeout: float = None
    ) -> None:
        """
        在指定 session 上下文中运行协程。异常将会抛出
        """
        if session._handler is None:
            BotSessionManager.inject(session, self)
        try:
            s_token = SESSION_LOCAL._add_ctx(session)
            await aio.wait_for(aio.create_task(coro), timeout=timeout)
        finally:
            SESSION_LOCAL._del_ctx(s_token)
            # 这里可能因 bot 停止运行，导致退出事件执行方法时 session 为挂起态。因此需要强制唤醒
            if session._hup_signal.is_set():
                BotSessionManager._rouse(session)

    async def _run(self, event: MsgEvent) -> None:
        """
        获取 session 然后准备运行 executor
        """
        try:
            session = None
            if not self._direct_rouse:
                res = await BotSessionManager.try_attach(event, self)
                if res:
                    return
            session = await BotSessionManager.get(event, self.responder, self)
            # 如果因为冲突没有获得 session，且指定了冲突回调
            if session is None and self._conflict_cb_maker:
                temp_session = BotSessionManager.make_temp(event, self.responder)
                await self._run_on_ctx(self._conflict_cb_maker(), temp_session)
            # 如果因为冲突没有获得 session，但没有冲突回调
            if session is None:
                return
            # 如果没有冲突，正常获得到了 session
            try:
                exec_coro = self.executor(self._plugin)
                await self._run_on_ctx(exec_coro, session, self.timeout)
            except aio.TimeoutError:
                if self._overtime_cb_maker:
                    await self._run_on_ctx(self._overtime_cb_maker(), session)
        except DirectRetSignal:
            pass
        except Exception as e:
            e_name = e.__class__.__name__
            executor_name = self.executor.__qualname__
            self.logger.error(
                f"插件 {self._plugin.__class__.__id__} 事件处理方法 {executor_name} 发生异常：[{e_name}] {e}"
            )
            self.logger.debug(f"异常点的事件记录为：{event.raw}")
            self.logger.debug("异常回溯栈：\n" + traceback.format_exc().strip("\n"))
        finally:
            if session is None:
                return
            BotSessionManager.recycle(session, alive=self._hold)

    async def evoke(self, event: MsgEvent) -> bool:
        """
        接收总线分发的事件的方法。返回是否决定处理的判断。
        便于 disptacher 进行优先级阻断。校验通过会自动处理事件。
        """
        if not self._verify(event):
            return False

        if self._direct_rouse:
            res = await BotSessionManager.try_attach(event, self)
            if res:
                return True

        if not self.is_valid:
            return False

        if not self.temp:
            aio.create_task(self._run(event))
            return True

        async with self._run_lock:
            if self.is_valid:
                aio.create_task(self._run(event))
                self._invalidate()
                return True
            else:
                return False


class MsgEventHandler(EventHandler):
    def __init__(
        self,
        executor: AsyncFunc[None],
        plugin: Any,
        responder: IActionResponder,
        logger: Logger,
        matcher: BotMatcher = None,
        parser: BotParser = None,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: float = None,
        set_block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb_maker: Callable[[None], Coroutine] = None,
        overtime_cb_maker: Callable[[None], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            timeout,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb_maker,
            overtime_cb_maker,
        )
        self.matcher = matcher
        self.parser = parser

        # matcher 和 parser 不能同时存在
        if self.matcher and self.parser:
            raise EventHandlerError("参数 matcher 和 parser 不能同时存在")

    def _match_and_parse(self, event: MsgEvent) -> bool:
        """
        通过验权后，尝试对事件进行匹配。
        有普通的匹配器（matcher）匹配，也可以使用解析器（parser）匹配
        """
        if self.matcher:
            return self.matcher.match(event.text)
        if self.parser:
            args_group = event._get_args(self.parser.id)
            if args_group == -1:
                args_group = self.parser.parse(event.text)
                event._store_args(self.parser.id, args_group)
            res, cmd_name, args = self.parser.test(args_group)
            if not res:
                return False
            try:
                if self.parser.need_format:
                    self.parser.format(args)
                return True
            except ArgFormatFailed as e:
                msg = f"命令 {cmd_name} 参数格式化失败：\n" + e.__str__()
                action = msg_action(
                    msg, event.is_private(), event.sender.id, event.group_id
                )
                aio.create_task(self.responder.take_action(action))
                return False
        return True

    async def evoke(self, event: MsgEvent) -> bool:
        """
        接收总线分发的事件的方法。返回是否决定处理的判断。
        便于 disptacher 进行优先级阻断。校验通过会自动处理事件。
        """
        if not self._verify(event):
            return False

        if self._direct_rouse:
            res = await BotSessionManager.try_attach(event, self)
            if res:
                return True

        if not self.is_valid:
            return False

        if not self._match_and_parse(event):
            return False

        if not self.temp:
            aio.create_task(self._run(event))
            return True

        async with self._run_lock:
            if self.is_valid:
                aio.create_task(self._run(event))
                self._invalidate()
                return True
            else:
                return False


class ReqEventHandler(EventHandler):
    def __init__(
        self,
        executor: AsyncFunc[None],
        plugin: Any,
        responder: IActionResponder,
        logger: Logger,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: float = None,
        set_block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb_maker: Callable[[None], Coroutine] = None,
        overtime_cb_maker: Callable[[None], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            timeout,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb_maker,
            overtime_cb_maker,
        )


class NoticeEventHandler(EventHandler):
    def __init__(
        self,
        executor: AsyncFunc[None],
        plugin: Any,
        responder: IActionResponder,
        logger: Logger,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: float = None,
        set_block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb_maker: Callable[[None], Coroutine] = None,
        overtime_cb_maker: Callable[[None], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            timeout,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb_maker,
            overtime_cb_maker,
        )


class MetaEventHandler(EventHandler):
    def __init__(
        self,
        executor: AsyncFunc[None],
        plugin: Any,
        responder: IActionResponder,
        logger: Logger,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: float = None,
        set_block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb_maker: Callable[[None], Coroutine] = None,
        overtime_cb_maker: Callable[[None], Coroutine] = None,
    ) -> None:
        super().__init__(
            executor,
            plugin,
            responder,
            logger,
            checker,
            priority,
            timeout,
            set_block,
            temp,
            session_rule,
            session_hold,
            direct_rouse,
            conflict_wait,
            conflict_cb_maker,
            overtime_cb_maker,
        )


# 事件方法（事件执行器）构造参数
EventHandlerArgs = NamedTuple(
    "EventHandlerArgs", executor=AsyncFunc[None], type=EventHandler, params=List[Any]
)
